
#include <stdio.h>
#include <math.h>
#include <varargs.h>

#define END -1

typedef struct COMP {
    double re, im;
} complex_struct;

complex_struct *complex_stack;	/* a pointer to the beginning of the stack */
typedef int complex;		/* a complex is just an index in the stack */
int complex_size;		/* max. number of elements in current stack */
complex complex_current;	/* last used position in the stack */

/*	getting from a handle to a pointer to a struct */

#define COMPLEX(x) (*(complex_stack + (x)))
#define RE(x) (COMPLEX(x).re)
#define IM(x) (COMPLEX(x).im)

int complex_initialize(int size)	/* call before using the stack */
{
    complex_stack = (complex_struct *) malloc(size*sizeof(complex_struct));
    complex_current = 0;
    complex_size = size;
    return (int) complex_stack;
}

complex get_temp_complex()	/* get a new (temporary) complex number */
{
    if(complex_current == complex_size)
	{
	    complex_size *= 2;
	    complex_stack = 
		(complex_struct *) realloc(complex_stack,
					   complex_size*sizeof(complex_struct));
	}
    if(!complex_stack)
	{
	    fprintf(stderr, "Out of memory in allocating a new complex.\n");
	    return(-1);
	}
    return(complex_current++);
}

complex new_complex( double r, double i)	/* allocate a complex, and
						   set its value */
{
    complex temp = get_temp_complex();

    if(temp != -1)
	{
	    RE(temp) = r;
	    IM(temp) = i;
	}
    return temp;
}

complex copy_complex( complex z)
{
    return new_complex( RE(z), IM(z) );
}

void complex_restart()
{
    complex_current = 0;
}

void complex_terminate()
{
    free( complex_stack );
}

double cabs_squared( complex z )
{
    return RE(z)*RE(z) + IM(z)*IM(z);
}

double Cabs( complex z )
{
    return sqrt(RE(z) * RE(z) + IM(z) * IM(z));
}

complex conjugate( complex z )
{
    return new_complex( RE(z), -IM(z) );
}

complex cexp( complex z)
{
    double d = exp(RE(z));
    return new_complex( d*cos(IM(z)), d*sin(IM(z)));
}

double carg( complex z )
{
    return atan2( IM(z), RE(z) );
}

complex clog( complex z )
{
    return new_complex( log(Cabs(z)), carg(z));
}

complex cpower( complex z1, complex z2)
{
    return cexp( ctimes( z2, clog( z1 ), END ) );
}

complex csqrt( complex z )
{
    double d = sqrt( Cabs ( z ) ), a = carg( z ) / 2;
    return new_complex( d * cos(a), d * sin(a) );
}
complex dtoc(double x)
{
    return new_complex( x, 0.0 );
}

void add_to( complex z1, complex z2 )
{
    RE(z1) += RE(z2);
    IM(z1) += IM(z2);
}

void mult_by_complex(complex z1, complex z2 )
{
    double temp = RE(z1);

    RE(z1) = temp * RE(z2) - IM(z1) * IM(z2);
    IM(z1) = temp * IM(z2) + IM(z1) * RE(z2);
}

complex mult_by_double( complex z, double c )
{
    return new_complex(c*RE(z), c*IM(z));
}
    
complex cplus(va_alist)
va_dcl
{
    va_list ap;
    complex c;
    complex temp = get_temp_complex();

    va_start(ap);
    COMPLEX(temp) = COMPLEX( c = va_arg(ap, complex) );
    if( c == END ) return new_complex( 0.0, 0.0 );
    while( ( c = va_arg(ap, complex) ) != END)
        	add_to( temp, c );
    va_end(ap);

    return( temp );
}

complex ctimes(va_alist)
va_dcl
{
    va_list ap;
    complex c;
    complex temp = get_temp_complex();

    va_start(ap);
    COMPLEX(temp) = COMPLEX( c = va_arg(ap, complex ));
    if( c == END ) return new_complex( 1.0, 0.0 );
    while( ( c = va_arg(ap, complex) ) != END )
	mult_by_complex( temp, c );
    va_end(ap);

    return( temp );
}

void cfprint( char *file_name, complex z )
{
    fprintf( file_name, "(%g, %g)", RE(z), IM(z) );
}

void cprint( complex z )
{
    printf( "(%g, %g)", RE(z), IM(z) );
}
