
#include <stdio.h>
#include <math.h>
#include "complex.h"

typedef struct FOUR {
    double x, y, z, t;
} fv_struct;

fv_struct *fv_stack;	/* a pointer to the beginning of the stack */
typedef int fv;		/* a fv is just an index in the stack */
int fv_size;		/* max. number of elements in current stack */
fv fv_current;	/* last used position in the stack */

/*	getting from a handle to a pointer to a struct */

#define FV(x) (fv_stack + (x))
#define DFV(x, i) ((double *) FV(x))[i]
#define X(p) (FV(p)->x)
#define Y(p) (FV(p)->y)
#define Z(p) (FV(p)->z)
#define T(p) (FV(p)->t)

int fv_initialize(int size)	/* call before using the stack */
{
    fv_stack = (fv_struct *) malloc(size*sizeof(fv_struct));
    fv_current = 0;
    fv_size = size;
    return (int) fv_stack;
}

fv get_temp_fv()	/* get a new (temporary) fv number */
{
    if(fv_current == fv_size)
	{
	    fv_size *= 2;
	    fv_stack = 
		(fv_struct *) realloc(fv_stack,
					   fv_size*sizeof(fv_struct));
	}
    if(!fv_stack)
	{
	    fprintf(stderr, "Out of memory in allocating a new fv.\n");
	    return(-1);
	}
    return(fv_current++);
}

fv new_fv( double x, double y, double z, double t) /* allocate a fv, and
						   set its value */
{
    fv temp = get_temp_fv();

    if(temp != -1)
	{
	    X(temp) = x;
	    Y(temp) = y;
	    Z(temp) = z;
	    T(temp) = t;
	}
    return temp;
}

fv copy_fv( fv p )
{
    return new_fv( X(p), Y(p), Z(p), T(p) );
}

void fv_restart()
{
    fv_current = 0;
}

void fv_terminate()
{
    free( fv_stack );
}

fv fv_sum(fv fv1, fv fv2)
{
    return new_fv(X(fv1)+X(fv2), Y(fv1)+Y(fv2), Z(fv1)+Z(fv2), T(fv1)+T(fv2));
}

fv fv_difference(fv fv1, fv fv2)
{
    return new_fv(X(fv1)-X(fv2), Y(fv1)-Y(fv2), Z(fv1)-Z(fv2), T(fv1)-T(fv2));
}

double dot_product(fv p1, fv p2)
{
    return T(p1)*T(p2)-X(p1)*X(p2)-Y(p1)*Y(p2)-Z(p1)*Z(p2);
}

void fv_fprint(FILE *fo, fv p)
{
    double theta, cth, phi, xm, ap = sqrt(X(p)*X(p)+Y(p)*Y(p)+Z(p)*Z(p)), 
    pp = sqrt(X(p)*X(p)+Y(p)*Y(p)),
    xms = dot_product(p, p);
    xm = ( xms < 0.0 ) ? -sqrt( -xms ) : sqrt( xms );
    fprintf(fo, "= (%g, %g, %g, %g) -> %g\n", X(p), Y(p), Z(p), T(p), xm);
    fprintf(fo, "|p| = %g, PP = %g, ", ap, pp);
    if( ap < 0.001 )
	fprintf(fo, "theta and phi are undefined.\n");
    else
	{
	    cth = Z(p) / ap;
	    theta = acos(cth) * 180.0 / M_PI;
	    fprintf(fo, "theta = %g, ", theta);
	    if( abs(cth) > 0.98 )
		fprintf(fo, "phi is undefined.\n");
	    else
		{
		    phi = atan2( Y(p), X(p) );
		    fprintf(fo, "phi = %g.\n", phi);
		}
	}
}

void fv_print(fv p)
{
    fv_fprint(stdout, p);
    return;
}

complex spinor_s(fv p, fv q)
{
    double pp = T(p) + X(p), pm = T(p) - X(p);
    double qp = T(q) + X(q), qm = T(q) - X(q);
    double pbot, qbot;
    complex ppb = new_complex( Y(p), -Z(p) );
    complex qqb = new_complex( Y(q), -Z(q) );

    if( pp < 1.0e-8 ) pp = 0.0;
    if( pm < 1.0e-8 ) pm = 0.0;
    if( qp < 1.0e-8 ) qp = 0.0;
    if( qm < 1.0e-8 ) qm = 0.0;
    if(( pbot = Cabs( ppb ) ) < 1.0e-10 * T(p) ) 
	ppb = dtoc( 1.0 );
    else
	ppb = mult_by_double( ppb, 1/pbot );
    if(( qbot = Cabs( qqb ) ) < 1.0e-10 * T(q) ) 
	qqb = dtoc( 1.0 );
    else
	qqb = mult_by_double( qqb, 1/qbot );

    return cplus(mult_by_double(ppb, sqrt(qp*pm)), 
		 mult_by_double(qqb, -sqrt(pp*qm)), END);
}

complex spinor_t(fv p, fv q)
{
    return conjugate( spinor_s( q, p ) );
}

double cosine(fv p, fv q)
{
    double a1 = X(p)*X(p) + Y(p)*Y(p) + Z(p)*Z(p),
    a2 = X(q)*X(q) + Y(q)*Y(q) + Z(q)*Z(q);
    
    if( a1*a2 < 1.0e-10 ) return( 1.0 );

    return (X(p)*X(q) + Y(p)*Y(q) + Z(p)*Z(q)) / sqrt( a1*a2 );
}

double perpp(fv p)
{
    return sqrt( X(p)*X(p) + Y(p)*Y(p) );
}

fv opposite(fv p)
{
    return new_fv( -X(p), -Y(p), -Z(p), sqrt( X(p)*X(p)+Y(p)*Y(p)+Z(p)*Z(p) ));
}

/*
void new_direction(double cth, double phi, double zcth, double zphi,
		   double *ncth, double *nphi)
{
    double x, y, z;

    x = sqrt(1 - ct*ct) * ct1 * cos(p) + sqrt(1 - ct1 * ct1) * cos(p1) - 
    sqrt(1 - ct1*ct1) cos(p)*cos(p) * cos(p1) + 
    ct * sqrt(1 - ct1*ct1) * cos(p)*cos(p) * cos(p1) - 
    sqrt(1 - ct1*ct1) * cos(p) * sin(p) * sin(p1) + 
    ct * sqrt(1 - ct1*ct1) * cos(p) * sin(p) * sin(p1);

    y = Sqrt(1 - Power(ct,2))*ct1*Sin(p) - 
    Sqrt(1 - Power(ct1,2))*Cos(p)*Cos(p1)*Sin(p) + 
    ct*Sqrt(1 - Power(ct1,2))*Cos(p)*Cos(p1)*Sin(p) + 
    ct*Sqrt(1 - Power(ct1,2))*Sin(p1) + 
    Sqrt(1 - Power(ct1,2))*Power(Cos(p),2)*Sin(p1) - 
    ct*Sqrt(1 - Power(ct1,2))*Power(Cos(p),2)*Sin(p1);

    z = ct*ct1 - Sqrt(1 - Power(ct,2))*Sqrt(1 - Power(ct1,2))*Cos(p)*
     Cos(p1) - Sqrt(1 - Power(ct,2))*Sqrt(1 - Power(ct1,2))*Sin(p)*Sin(p1);
}
*/

double mass(fv p)
{
    double ms = T(p)*T(p)-Z(p)*Z(p)-Y(p)*Y(p)-X(p)*X(p);

    return (ms < 0.0) ? -sqrt(-ms) : sqrt(ms);
}

double phase_space_factor(double m, double m1, double m2)
{
    double ms = m*m, ms1 = m1*m1, ms2 = m2*m2;

    return 
	sqrt(ms*ms + ms1*ms1 + ms2*ms2 - 2*ms*ms1 - 2*ms*ms2 - 2*ms1*ms2)/ms;
}

void decay(fv p0, double cth, double phi, double m1, double m2, fv *r1, fv *r2,
	   double *zaxis)
{
    double m = mass(p0), dot, p, e1 = (m*m + m1*m1 - m2*m2)/(2*m);
    double sth = sqrt(1 - cth*cth), len, n[3];
    int i;

    if(!zaxis || (zaxis[0] == 0 && zaxis[1] == 0))
	{
	    n[0] = sth*cos(phi);
	    n[1] = sth*sin(phi);
	    n[2] = cth;
	}
    else
	{
	    len = sqrt(zaxis[0]*zaxis[0] + zaxis[1]*zaxis[1] + 
		       zaxis[2]*zaxis[2]);
	    n[0] = (cth*zaxis[0]*zaxis[0]*zaxis[0] - 
		    len*sth*sin(phi)*zaxis[0]*zaxis[1] +
		    len*sth*cos(phi)*zaxis[1]*zaxis[1] +
		    cth*zaxis[0]*zaxis[1]*zaxis[1] +
		    sth*cos(phi)*zaxis[0]*zaxis[0]*zaxis[2] +
		    sth*sin(phi)*zaxis[0]*zaxis[1]*zaxis[2]) /
			(len*(zaxis[0]*zaxis[0] + zaxis[1]*zaxis[1])); 
	    n[1] = (len*sth*sin(phi)*zaxis[0]*zaxis[0] -
		    len*sth*cos(phi)*zaxis[0]*zaxis[1] + 
		    cth*zaxis[0]*zaxis[0]*zaxis[1] +
		    cth*zaxis[1]*zaxis[1]*zaxis[1] + 
		    sth*cos(phi)*zaxis[0]*zaxis[1]*zaxis[2] +
		    sth*sin(phi)*zaxis[1]*zaxis[1]*zaxis[2]) /
			(len*(zaxis[0]*zaxis[0] + zaxis[1]*zaxis[1]));
	    n[2] = (-(sth*cos(phi)*zaxis[0]) - sth*sin(phi)*zaxis[1] +
		    cth*zaxis[2]) / len;
	}
    p = m * phase_space_factor(m, m1, m2) / 2;

    dot = DFV(p0, 0)*n[0] + DFV(p0, 1)*n[1] + DFV(p0, 2)*n[2];
    *r1 = new_fv(0.0, 0.0, 0.0, 0.0);
    *r2 = new_fv(0.0, 0.0, 0.0, 0.0);
    for ( i = 0; i <= 2; i++)
	DFV(*r1, i) = p * ( n[i] + DFV(p0, i) * dot / (m * (m + DFV(p0, 3)))) +
	    DFV(p0, i) * e1/m;
    DFV(*r1, 3) = DFV(p0, 3) * e1/m + dot * p / m;

    for ( i = 0; i <= 3; i++)
	DFV(*r2, i) = DFV(p0, i) - DFV(*r1, i);
}

/*
void decay(fv p, double cth, double phi, double m1, double m2, fv *p1,
	   fv *p2, double zcth, double zphi)
{
    double ncth, nphi;

    new_direction(cth, phi, zcth, zphi, &ncth, &nphi);
}
*/

fv boost(fv p, double x)
{
  return new_fv(X(p), Y(p), Z(p)*cosh(x)+T(p)*sinh(x), 
		Z(p)*sinh(x)+T(p)*cosh(x));
}
